@@ -5,7 +5,7 @@
 # Written 91-12-02 through 92-01-01 by Stephen McGee.
 # Modified 92-02-11 through 92-02-22 by Chris Arthur to further generalize.
 #
-# Copyright (C) 1991-2013 Free Software Foundation, Inc.
+# Copyright (C) 1991-2019 Free Software Foundation, Inc.
 # This file is part of GNU Make.
 #
 # GNU Make is free software; you can redistribute it and/or modify it under
@@ -48,12 +48,55 @@ $tests_passed = 0;
 # Yeesh.  This whole test environment is such a hack!
 $test_passed = 1;
 
-
 # Timeout in seconds.  If the test takes longer than this we'll fail it.
 $test_timeout = 5;
+$test_timeout = 10 if $^O eq 'VMS';
 
-# Path to Perl
+# Path to Perl--make sure it uses forward-slashes
 $perl_name = $^X;
+$perl_name =~ tr,\\,/,;
+
+# Find the strings that will be generated for various error codes.
+# We want them from the C locale regardless of our current locale.
+
+my $loc = undef;
+if ($has_POSIX) {
+    $loc = POSIX::setlocale(POSIX::LC_MESSAGES);
+    POSIX::setlocale(POSIX::LC_MESSAGES, 'C');
+}
+
+$ERR_no_such_file = undef;
+$ERR_read_only_file = undef;
+$ERR_unreadable_file = undef;
+
+if (open(my $F, '<', 'file.none')) {
+    print "Opened non-existent file! Skipping related tests.\n";
+} else {
+    $ERR_no_such_file = "$!";
+}
+
+unlink('file.out');
+touch('file.out');
+
+chmod(0444, 'file.out');
+if (open(my $F, '>', 'file.out')) {
+    print "Opened read-only file! Skipping related tests.\n";
+    close($F);
+} else {
+    $ERR_read_only_file = "$!";
+}
+
+chmod(0000, 'file.out');
+if (open(my $F, '<', 'file.out')) {
+    print "Opened unreadable file!  Skipping related tests.\n";
+    close($F);
+} else {
+    $ERR_unreadable_file = "$!";
+}
+
+unlink('file.out') or die "Failed to delete file.out: $!\n";
+
+$loc and POSIX::setlocale(POSIX::LC_MESSAGES, $loc);
 
 # %makeENV is the cleaned-out environment.
 %makeENV = ();
@@ -62,19 +105,67 @@ $perl_name = $^X;
 # These are RESET AFTER EVERY TEST!
 %extraENV = ();
 
+sub vms_get_process_logicals {
+  # Sorry for the long note here, but to keep this test running on
+  # VMS, it is needed to be understood.
+  #
+  # Perl on VMS by default maps the %ENV array to the system wide logical
+  # name table.
+  #
+  # This is a very large dynamically changing table.
+  # On Linux, this would be the equivalent of a table that contained
+  # every mount point, temporary pipe, and symbolic link on every
+  # file system.  You normally do not have permission to clear or replace it,
+  # and if you did, the results would be catastrophic.
+  #
+  # On VMS, added/changed %ENV items show up in the process logical
+  # name table.  So to track changes, a copy of it needs to be captured.
+
+  my $raw_output = `show log/process/access_mode=supervisor`;
+  my @raw_output_lines = split('\n',$raw_output);
+  my %log_hash;
+  foreach my $line (@raw_output_lines) {
+    if ($line =~ /^\s+"([A-Za-z\$_]+)"\s+=\s+"(.+)"$/) {
+      $log_hash{$1} = $2;
+    }
+  }
+  return \%log_hash
+}
+
 # %origENV is the caller's original environment
-%origENV = %ENV;
+if ($^O ne 'VMS') {
+  %origENV = %ENV;
+} else {
+  my $proc_env = vms_get_process_logicals;
+  %origENV = %{$proc_env};
+}
 
 sub resetENV
 {
   # We used to say "%ENV = ();" but this doesn't work in Perl 5.000
   # through Perl 5.004.  It was fixed in Perl 5.004_01, but we don't
   # want to require that here, so just delete each one individually.
-  foreach $v (keys %ENV) {
-    delete $ENV{$v};
+
+  if ($^O ne 'VMS') {
+    foreach $v (keys %ENV) {
+      delete $ENV{$v};
+    }
+
+    %ENV = %makeENV;
+  } else {
+    my $proc_env = vms_get_process_logicals();
+    my %delta = %{$proc_env};
+    foreach my $v (keys %delta) {
+      if (exists $origENV{$v}) {
+        if ($origENV{$v} ne $delta{$v}) {
+          $ENV{$v} = $origENV{$v};
+        }
+      } else {
+        delete $ENV{$v};
+      }
+    }
   }
 
-  %ENV = %makeENV;
   foreach $v (keys %extraENV) {
     $ENV{$v} = $extraENV{$v};
     delete $extraENV{$v};
@@ -88,6 +179,8 @@ sub toplevel
   foreach (# UNIX-specific things
            'TZ', 'TMPDIR', 'HOME', 'USER', 'LOGNAME', 'PATH',
            'LD_LIBRARY_PATH',
+           # SAN things
+           'ASAN_OPTIONS', 'UBSAN_OPTIONS',
            # Purify things
            'PURIFYOPTIONS',
            # Windows NT-specific stuff
@@ -105,7 +198,7 @@ sub toplevel
 
   # Replace the environment with the new one
   #
-  %origENV = %ENV;
+  %origENV = %ENV unless $^O eq 'VMS';
 
   resetENV();
 
@@ -139,6 +232,25 @@ sub toplevel
 
   &print_banner;
 
+  if ($osname eq 'VMS' && $cwdslash eq "")
+  {
+    # Porting this script to VMS revealed a small bug in opendir() not
+    # handling search lists correctly when the directory only exists in
+    # one of the logical_devices.  Need to find the first directory in
+    # the search list, as that is where things will be written to.
+    my @dirs = split("/", $pwd);
+
+    my $logical_device = $ENV{$dirs[1]};
+    if ($logical_device =~ /([A-Za-z0-9_]+):(:?.+:)+/)
+    {
+        # A search list was found.  Grab the first logical device
+        # and use it instead of the search list.
+        $dirs[1]=$1;
+        my $lcl_pwd = join('/', @dirs);
+        $workpath = $lcl_pwd . '/' . $workdir
+    }
+  }
+
   if (-d $workpath)
   {
     print "Clearing $workpath...\n";
@@ -181,8 +293,10 @@ sub toplevel
     {
       next if ($dir =~ /^(\..*|CVS|RCS)$/ || ! -d "$scriptpath/$dir");
       push (@rmdirs, $dir);
+      # VMS can have overlayed file systems, so directories may repeat.
+      next if -d "$workpath/$dir";
       mkdir ("$workpath/$dir", 0777)
-           || &error ("Couldn't mkdir $workpath/$dir: $!\n");
+          || &error ("Couldn't mkdir $workpath/$dir: $!\n");
       opendir (SCRIPTDIR, "$scriptpath/$dir")
           || &error ("Couldn't opendir $scriptpath/$dir: $!\n");
       @files = grep (!/^(\..*|CVS|RCS|.*~)$/, readdir (SCRIPTDIR) );
@@ -239,6 +353,47 @@ sub get_osname
   # Set up an initial value.  In perl5 we can do it the easy way.
   $osname = defined($^O) ? $^O : '';
 
+  # find the type of the port.  We do this up front to have a single
+  # point of change if it needs to be tweaked.
+  #
+  # This is probably not specific enough.
+  #
+  if ($osname =~ /MSWin32/i || $osname =~ /Windows/i
+      || $osname =~ /MINGW32/i || $osname =~ /CYGWIN_NT/i) {
+    $port_type = 'W32';
+  }
+  # Bleah, the osname is so variable on DOS.  This kind of bites.
+  # Well, as far as I can tell if we check for some text at the
+  # beginning of the line with either no spaces or a single space, then
+  # a D, then either "OS", "os", or "ev" and a space.  That should
+  # match and be pretty specific.
+  elsif ($osname =~ /^([^ ]*|[^ ]* [^ ]*)D(OS|os|ev) /) {
+    $port_type = 'DOS';
+  }
+  # Check for OS/2
+  elsif ($osname =~ m%OS/2%) {
+    $port_type = 'OS/2';
+  }
+
+  # VMS has a GNV Unix mode or a DCL mode.
+  # The SHELL environment variable should not be defined in VMS-DCL mode.
+  elsif ($osname eq 'VMS' && !defined $ENV{"SHELL"}) {
+    $port_type = 'VMS-DCL';
+  }
+  # Everything else, right now, is UNIX.  Note that we should integrate
+  # the VOS support into this as well and get rid of $vos; we'll do
+  # that next time.
+  else {
+    $port_type = 'UNIX';
+  }
+
+  if ($osname eq 'VMS')
+  {
+    $vos = 0;
+    $pathsep = "/";
+    return;
+  }
+
   # Find a path to Perl
 
   # See if the filesystem supports long file names with multiple
@@ -419,7 +574,11 @@ sub run_all_tests
 {
     $categories_run = 0;
 
+    $lasttest = '';
     foreach $testname (sort @TESTS) {
+        # Skip duplicates on VMS caused by logical name search lists.
+        next if $testname eq $lasttest;
+        $lasttest = $testname;
         $suite_passed = 1;       # reset by test on failure
         $num_of_logfiles = 0;
         $num_of_tmpfiles = 0;
@@ -445,12 +604,15 @@ sub run_all_tests
             $runext = 'run';
             $extext = '.';
         }
+        $extext = '_' if $^O eq 'VMS';
         $log_filename = "$testpath.$logext";
         $diff_filename = "$testpath.$diffext";
         $base_filename = "$testpath.$baseext";
         $run_filename = "$testpath.$runext";
         $tmp_filename = "$testpath.$tmpfilesuffix";
 
+        -f $perl_testname or die "Invalid test: $testname\n\n";
+
         setup_for_test();
 
         $output = "........................................................ ";
@@ -686,6 +848,121 @@ sub compare_output
           $slurp_mod =~ s,\r\n,\n,gs;
 
           $answer_matched = ($slurp_mod eq $answer_mod);
+          if ($^O eq 'VMS') {
+
+            # VMS has extra blank lines in output sometimes.
+            # Ticket #41760
+            if (!$answer_matched) {
+              $slurp_mod =~ s/\n\n+/\n/gm;
+              $slurp_mod =~ s/\A\n+//g;
+              $answer_matched = ($slurp_mod eq $answer_mod);
+            }
+
+            # VMS adding a "Waiting for unfinished jobs..."
+            # Remove it for now to see what else is going on.
+            if (!$answer_matched) {
+              $slurp_mod =~ s/^.+\*\*\* Waiting for unfinished jobs.+$//m;
+              $slurp_mod =~ s/\n\n/\n/gm;
+              $slurp_mod =~ s/^\n+//gm;
+              $answer_matched = ($slurp_mod eq $answer_mod);
+            }
+
+            # VMS wants target device to exist or generates an error,
+            # Some test tagets look like VMS devices and trip this.
+            if (!$answer_matched) {
+              $slurp_mod =~ s/^.+\: no such device or address.*$//gim;
+              $slurp_mod =~ s/\n\n/\n/gm;
+              $slurp_mod =~ s/^\n+//gm;
+              $answer_matched = ($slurp_mod eq $answer_mod);
+            }
+
+            # VMS error message has a different case
+            if (!$answer_matched) {
+              $slurp_mod =~ s/no such file /No such file /gm;
+              $answer_matched = ($slurp_mod eq $answer_mod);
+            }
+
+            # VMS is putting comas instead of spaces in output
+            if (!$answer_matched) {
+              $slurp_mod =~ s/,/ /gm;
+              $answer_matched = ($slurp_mod eq $answer_mod);
+            }
+
+            # VMS Is sometimes adding extra leading spaces to output?
+            if (!$answer_matched) {
+               my $slurp_mod = $slurp_mod;
+               $slurp_mod =~ s/^ +//gm;
+               $answer_matched = ($slurp_mod eq $answer_mod);
+            }
+
+            # VMS port not handling POSIX encoded child status
+            # Translate error case it for now.
+            if (!$answer_matched) {
+              $slurp_mod =~ s/0x1035a00a/1/gim;
+              $answer_matched = 1 if $slurp_mod =~ /\Q$answer_mod\E/i;
+
+            }
+            if (!$answer_matched) {
+              $slurp_mod =~ s/0x1035a012/2/gim;
+              $answer_matched = ($slurp_mod eq $answer_mod);
+            }
+
+            # Tests are using a UNIX null command, temp hack
+            # until this can be handled by the VMS port.
+            # ticket # 41761
+            if (!$answer_matched) {
+              $slurp_mod =~ s/^.+DCL-W-NOCOMD.*$//gim;
+              $slurp_mod =~ s/\n\n+/\n/gm;
+              $slurp_mod =~ s/^\n+//gm;
+              $answer_matched = ($slurp_mod eq $answer_mod);
+            }
+            # Tests are using exit 0;
+            # this generates a warning that should stop the make, but does not
+            if (!$answer_matched) {
+              $slurp_mod =~ s/^.+NONAME-W-NOMSG.*$//gim;
+              $slurp_mod =~ s/\n\n+/\n/gm;
+              $slurp_mod =~ s/^\n+//gm;
+              $answer_matched = ($slurp_mod eq $answer_mod);
+            }
+
+            # VMS is sometimes adding single quotes to output?
+            if (!$answer_matched) {
+              my $noq_slurp_mod = $slurp_mod;
+              $noq_slurp_mod =~ s/\'//gm;
+              $answer_matched = ($noq_slurp_mod eq $answer_mod);
+
+              # And missing an extra space in output
+              if (!$answer_matched) {
+                $noq_answer_mod = $answer_mod;
+                $noq_answer_mod =~ s/\h\h+/ /gm;
+                $answer_matched = ($noq_slurp_mod eq $noq_answer_mod);
+              }
+
+              # VMS adding ; to end of some lines.
+              if (!$answer_matched) {
+                $noq_slurp_mod =~ s/;\n/\n/gm;
+                $answer_matched = ($noq_slurp_mod eq $noq_answer_mod);
+              }
+
+              # VMS adding trailing space to end of some quoted lines.
+              if (!$answer_matched) {
+                $noq_slurp_mod =~ s/\h+\n/\n/gm;
+                $answer_matched = ($noq_slurp_mod eq $noq_answer_mod);
+              }
+
+              # And VMS missing leading blank line
+              if (!$answer_matched) {
+                $noq_answer_mod =~ s/\A\n//g;
+                $answer_matched = ($noq_slurp_mod eq $noq_answer_mod);
+              }
+
+              # Unix double quotes showing up as single quotes on VMS.
+              if (!$answer_matched) {
+                $noq_answer_mod =~ s/\"//g;
+                $answer_matched = ($noq_slurp_mod eq $noq_answer_mod);
+              }
+            }
+          }
 
           # If it still doesn't match, see if the answer might be a regex.
           if (!$answer_matched && $answer =~ m,^/(.+)/$,) {
@@ -781,40 +1058,99 @@ sub detach_default_output
   @OUTSTACK or error("default output stack has flown under!\n", 1);
 
   close(STDOUT);
-  close(STDERR);
+  close(STDERR) unless $^O eq 'VMS';
+
 
   open (STDOUT, '>&', pop @OUTSTACK) or error("ddo: $! duping STDOUT\n", 1);
   open (STDERR, '>&', pop @ERRSTACK) or error("ddo: $! duping STDERR\n", 1);
 }
 
+sub _run_with_timeout
+{
+    my $code;
+    if ($^O eq 'VMS') {
+        #local $SIG{ALRM} = sub {
+        #    my $e = $ERRSTACK[0];
+        #    print $e "\nTest timed out after $test_timeout seconds\n";
+        #    die "timeout\n";
+        #};
+        #alarm $test_timeout;
+        system(@_);
+        #alarm 0;
+        my $severity = ${^CHILD_ERROR_NATIVE} & 7;
+        $code = 0;
+        if (($severity & 1) == 0) {
+            $code = 512;
+        }
+
+        # Get the vms status.
+        my $vms_code = ${^CHILD_ERROR_NATIVE};
+
+        # Remove the print status bit
+        $vms_code &= ~0x10000000;
+
+        # Posix code translation.
+        if (($vms_code & 0xFFFFF000) == 0x35a000) {
+            $code = (($vms_code & 0xFFF) >> 3) * 256;
+        }
+
+    } elsif ($port_type eq 'W32') {
+        my $pid = system(1, @_);
+        $pid > 0 or die "Cannot execute $_[0]\n";
+        local $SIG{ALRM} = sub {
+            my $e = $ERRSTACK[0];
+            print $e "\nTest timed out after $test_timeout seconds\n";
+            kill -9, $pid;
+            die "timeout\n";
+        };
+        alarm $test_timeout;
+        my $r = waitpid($pid, 0);
+        alarm 0;
+        $r == -1 and die "No such pid: $pid\n";
+        # This shouldn't happen since we wait forever or timeout via SIGALRM
+        $r == 0 and die "No process exited.\n";
+        $code = $?;
+
+    } else {
+        my $pid = fork();
+        if (! $pid) {
+            exec(@_) or die "exec: Cannot execute $_[0]\n";
+        }
+        local $SIG{ALRM} = sub {
+            my $e = $ERRSTACK[0];
+            print $e "\nTest timed out after $test_timeout seconds\n";
+            # Resend the alarm to our process group to kill the children.
+            $SIG{ALRM} = 'IGNORE';
+            kill -14, $$;
+            die "timeout\n";
+        };
+        alarm $test_timeout;
+        my $r = waitpid($pid, 0);
+        alarm 0;
+        $r == -1 and die "No such pid: $pid\n";
+        # This shouldn't happen since we wait forever or timeout via SIGALRM
+        $r == 0 and die "No process exited.\n";
+        $code = $?;
+    }
+
+    return $code;
+}
+
 # This runs a command without any debugging info.
 sub _run_command
 {
-  my $code;
-
   # We reset this before every invocation.  On Windows I think there is only
   # one environment, not one per process, so I think that variables set in
   # test scripts might leak into subsequent tests if this isn't reset--???
   resetENV();
 
-  eval {
-      my $pid = fork();
-      if (! $pid) {
-          exec(@_) or die "Cannot execute $_[0]\n";
-      }
-      local $SIG{ALRM} = sub { my $e = $ERRSTACK[0]; print $e "\nTest timed out after $test_timeout seconds\n"; die "timeout\n"; };
-      alarm $test_timeout;
-      waitpid($pid, 0) > 0 or die "No such pid: $pid\n";
-      $code = $?;
-      alarm 0;
-  };
+  my $orig = $SIG{ALRM};
+  my $code = eval { _run_with_timeout(@_); };
+  $SIG{ALRM} = $orig;
+
   if ($@) {
       # The eval failed.  If it wasn't SIGALRM then die.
       $@ eq "timeout\n" or die "Command failed: $@";
-
-      # Timed out.  Resend the alarm to our process group to kill the children.
-      $SIG{ALRM} = 'IGNORE';
-      kill -14, $$;
       $code = 14;
   }
 
@@ -829,7 +1165,7 @@ sub run_command
   print "\nrun_command: @_\n" if $debug;
   my $code = _run_command(@_);
   print "run_command returned $code.\n" if $debug;
-
+  print "vms status = ${^CHILD_ERROR_NATIVE}\n" if $debug and $^O eq 'VMS';
   return $code;
 }
 
@@ -851,7 +1187,7 @@ sub run_command_with_output
   $err and die $err;
 
   print "run_command_with_output returned $code.\n" if $debug;
-
+  print "vms status = ${^CHILD_ERROR_NATIVE}\n" if $debug and $^O eq 'VMS';
   return $code;
 }
 
@@ -911,7 +1247,15 @@ sub remove_directory_tree_inner
     }
     else
     {
-      unlink $object || return 0;
+      if ($^O ne 'VMS')
+      {
+        unlink $object || return 0;
+      }
+      else
+      {
+        # VMS can have multiple versions of a file.
+        1 while unlink $object;
+      }
     }
   }
   closedir ($dirhandle);
